@@ -1,9 +1,10 @@
 import argparse
+import atexit
 import collections
 import copy
+import datetime
 import http.client
 import json
-import logging
 import os
 import getpass
 import re
@@ -14,8 +15,6 @@ import sys
 import tarfile
 import types
 
-import version
-
 
 ## Imports not in standard library ##
 
@@ -56,8 +55,14 @@ except ImportError:
 
 ## Globals ##
 
-# Verbosity level. Can be 0, 1, or 2.
-verbose = 0
+# FIXME: currently set in ch-grow :P
+CH_BIN = None
+CH_RUN = None
+
+# Logging; set using log_setup() below.
+verbose = 0          # Verbosity level. Can be 0, 1, or 2.
+log_festoon = False  # If true, prepend pid and timestamp to chatter.
+log_fp = sys.stderr  # File object to print logs to.
 
 # This is a general grammar for all the parsing we need to do. As such, you
 # must prepend a start rule before use.
@@ -127,16 +132,14 @@ OPTION_KEY: /[a-z]+/
 OPTION_VALUE: /[^ \t\n]+/
 
 HEX_STRING: /[0-9A-Fa-f]+/
-LINE: ( LINE_CONTINUE | /[^\n]/ )+
+LINE: ( _LINE_CONTINUE | /[^\n]/ )+
 WORD: /[^ \t\n=]/+
 
 _string_list: "[" _WS? STRING_QUOTED ( "," _WS? STRING_QUOTED )* _WS? "]"
 
-LINE_CONTINUE: "\\\n"
-%ignore LINE_CONTINUE
-
 _NEWLINES: _WS? "\n"+
-_WS: /[ \t]/+
+_WS: /[ \t]|\\\n/+
+_LINE_CONTINUE: "\\\n"
 
 %import common.ESCAPED_STRING -> STRING_QUOTED
 """
@@ -144,22 +147,29 @@ _WS: /[ \t]/+
 
 ## Classes ##
 
-class CLI_Action_Exit(argparse.Action):
+class HelpFormatter(argparse.HelpFormatter):
 
    def __init__(self, *args, **kwargs):
-      super().__init__(nargs=0, *args, **kwargs)
+      # max_help_position is undocumented but I don't know how else to do this.
+      #kwargs["max_help_position"] = 26
+      super().__init__(max_help_position=26, *args, **kwargs)
 
-class CLI_Dependencies(CLI_Action_Exit):
-
-   def __call__(self, *args, **kwargs):
-      dependencies_check()
-      sys.exit(0)
-
-class CLI_Version(CLI_Action_Exit):
-
-   def __call__(self, *args, **kwargs):
-      print(version.VERSION)
-      sys.exit(0)
+   # Suppress duplicate metavar printing when option has both short and long
+   # flavors. E.g., instead of:
+   #
+   #   -s DIR, --storage DIR  set builder internal storage directory to DIR
+   #
+   # print:
+   #
+   #   -s, --storage DIR      set builder internal storage directory to DIR
+   #
+   # From https://stackoverflow.com/a/31124505.
+   def _format_action_invocation(self, action):
+      if (not action.option_strings or action.nargs == 0):
+         return super()._format_action_invocation(action)
+      default = self._get_default_metavar_for_optional(action)
+      args_string = self._format_args(action, default)
+      return ', '.join(action.option_strings) + ' ' + args_string
 
 class Image:
    """Container image object.
@@ -214,9 +224,9 @@ class Image:
 
    def copy_unpacked(self, other):
       "Copy the unpack directory of Image other to my unpack directory."
-      DEBUG("copying image: %s -> %s" % (other.unpack_path, self.unpack_path))
       self.unpack_create_ok()
-      shutil.copytree(other.unpack_path, self.unpack_path, symlinks=True)
+      DEBUG("copying image: %s -> %s" % (other.unpack_path, self.unpack_path))
+      copytree(other.unpack_path, self.unpack_path, symlinks=True)
 
    def download(self, use_cache):
       """Download image manifest and layers according to origin and put them
@@ -250,27 +260,11 @@ class Image:
       "Add the Charliecloud workarounds to the unpacked image."
       DEBUG("fixing up image: %s" % self.unpack_path)
       # Metadata directory.
-      mkdirs("%s/ch/bin" % self.unpack_path)
+      mkdirs("%s/ch" % self.unpack_path)
       file_ensure_exists("%s/ch/environment" % self.unpack_path)
       # Mount points.
       file_ensure_exists("%s/etc/hosts" % self.unpack_path)
       file_ensure_exists("%s/etc/resolv.conf" % self.unpack_path)
-      # /etc/{passwd,group}
-      file_write("%s/etc/passwd" % self.unpack_path, """\
-root:x:0:0:root:/root:/bin/sh
-nobody:x:65534:65534:nobody:/:/bin/false
-""")
-      file_write("%s/etc/group" % self.unpack_path, """\
-root:x:0:
-nogroup:x:65534:
-""")
-      # Kludges to work around expectations of real root, not UID 0 in a
-      # unprivileged user namespace. See also the default environment.
-      #
-      # Debian "apt" and friends want to chown(1), chgrp(1), etc.
-      symlink("/bin/true", "%s/ch/bin/chown" % self.unpack_path)
-      symlink("/bin/true", "%s/ch/bin/chgrp" % self.unpack_path)
-      symlink("/bin/true", "%s/ch/bin/dpkg-statoverride" % self.unpack_path)
 
    def flatten(self):
       "Flatten the layers in the download cache into the unpack directory."
@@ -281,12 +275,15 @@ nogroup:x:65534:
       self.unpack_create()
       for (i, (lh, (fp, members))) in enumerate(layers.items(), start=1):
          INFO("layer %d/%d: %s: extracting" % (i, len(layers), lh[:7]))
-         fp.extractall(path=self.unpack_path, members=members)
+         try:
+            fp.extractall(path=self.unpack_path, members=members)
+         except OSError as x:
+            FATAL("can't extract layer %d: %s" % (i, x.strerror))
 
    def layer_hashes_load(self):
       "Load the layer hashes from the manifest file."
       try:
-         fp = open(self.manifest_path, "rt", encoding="UTF-8")
+         fp = open_(self.manifest_path, "rt", encoding="UTF-8")
       except OSError as x:
          FATAL("can't open manifest file: %s: %s"
                % (self.manifest_path, x.strerror))
@@ -448,9 +445,7 @@ nogroup:x:65534:
             FATAL("can't flatten: %s exists but does not appear to be an image"
                   % self.unpack_path)
          DEBUG("replacing existing image: %s" % self.unpack_path)
-         def fail(function, path, excinfo):
-            FATAL("can't flatten: %s: %s" % (path, excinfo[1]))
-         shutil.rmtree(self.unpack_path, onerror=fail)
+         rmtree(self.unpack_path)
 
    def unpack_create(self):
       "Ensure the unpack directory exists, replacing or creating if needed."
@@ -491,7 +486,6 @@ class Image_Ref:
    # first use.
    parser = None
 
-
    def __init__(self, src=None):
       self.host = None
       self.port = None
@@ -720,9 +714,9 @@ class Repo_Downloader:
       self.authenticate_maybe(url)
       res = self.get_raw(url, headers)
       try:
-         fp = open(path, "wb")
-         fp.write(res.content)
-         fp.close()
+         fp = open_(path, "wb")
+         ossafe(fp.write, "can't write: %s" % path, res.content)
+         ossafe(fp.close, "can't close: %s" % path)
       except OSError as x:
          FATAL("can't write: %s: %s" % (path, x))
 
@@ -786,18 +780,24 @@ class TarFile(tarfile.TarFile):
          [2]: https://bugs.python.org/issue19974"""
       try:
          st = os.lstat(targetpath)
+      except FileNotFoundError:
+         # We could move this except clause after all the stat.S_IS* calls,
+         # but that risks catching FileNotFoundError that came from somewhere
+         # other than lstat().
+         st = None
+      except OSError as x:
+         FATAL("can't lstat: %s" % targetpath, targetpath)
+      if (st is not None):
          if (stat.S_ISREG(st.st_mode)):
             pass  # regular file; do nothing (will be overwritten)
          elif (stat.S_ISDIR(st.st_mode)):
             FATAL("can't overwrite directory with regular file: %s"
                   % targetpath)
          elif (stat.S_ISLNK(st.st_mode)):
-            ossafe(os.unlink, "can't unlink: %s" % targetpath, targetpath)
+            unlink(targetpath)
          else:
             FATAL("invalid file type 0%o in previous layer; see inode(7): %s"
                   % (stat.S_IFMT(st.st_mode), targetpath))
-      except FileNotFoundError:
-         pass
       super().makefile(tarinfo, targetpath)
 
 
@@ -805,45 +805,50 @@ class TarFile(tarfile.TarFile):
 
 def DEBUG(*args, v=1, **kwargs):
    if (verbose >= v):
-      color("36m", sys.stderr)
-      print(flush=True, file=sys.stderr, *args, **kwargs)
-      color_reset(sys.stderr)
+      log(color="36m", *args, **kwargs)
 
 def ERROR(*args, **kwargs):
-   color("31m", sys.stderr)
-   print("error: ", file=sys.stderr, end="")
-   print(flush=True, file=sys.stderr, *args, **kwargs)
-   color_reset(sys.stderr)
+   log(color="31m", prefix="error: ", *args, **kwargs)
 
 def FATAL(*args, **kwargs):
    ERROR(*args, **kwargs)
    sys.exit(1)
 
 def INFO(*args, **kwargs):
-   print(flush=True, *args, **kwargs)
+   log(*args, **kwargs)
 
 def WARNING(*args, **kwargs):
-   color("31m", sys.stderr)
-   print("warning: ", file=sys.stderr, end="")
-   print(flush=True, file=sys.stderr, *args, **kwargs)
-   color_reset(sys.stderr)
+   log(color="31m", prefix="warning: ", *args, **kwargs)
+
+def ch_run_modify(img, args, env, workdir="/"):
+   args = [CH_BIN + "/ch-run", "-w", "--cd", workdir, "--uid=0", "--gid=0",
+           "--no-home", "--no-passwd", img, "--"] + args
+   cmd(args, env)
 
 def cmd(args, env=None):
    DEBUG("environment: %s" % env)
    DEBUG("executing: %s" % args)
-   color("33m", sys.stdout)
+   color_set("33m", sys.stdout)
    cp = subprocess.run(args, env=env, stdin=subprocess.DEVNULL)
    color_reset(sys.stdout)
    if (cp.returncode):
       FATAL("command failed with code %d: %s" % (cp.returncode, args[0]))
 
-def color(color, fp):
+def color_reset(*fps):
+   for fp in fps:
+      color_set("0m", fp)
+
+def color_set(color, fp):
    if (fp.isatty()):
       print("\033[" + color, end="", flush=True, file=fp)
 
-def color_reset(*fps):
-   for fp in fps:
-      color("0m", fp)
+def copy2(src, dst, **kwargs):
+   "Wrapper for shutil.copy2() with error checking."
+   ossafe(shutil.copy2, "can't copy: %s -> %s" % (src, dst), src, dst, **kwargs)
+
+def copytree(*args, **kwargs):
+   "Wrapper for shutil.copytree() that exits the program on the first error."
+   shutil.copytree(copy_function=copy2, *args, **kwargs)
 
 def dependencies_check():
    """Check more dependencies. If any dependency problems found, here or above
@@ -854,31 +859,84 @@ def dependencies_check():
       sys.exit(1)
 
 def file_ensure_exists(path):
-   with open(path, "a") as fp:
-      pass
+   fp = open_(path, "a")
+   fp.close()
 
 def file_write(path, content, mode=None):
-   with open(path, "wt") as fp:
-      fp.write(content)
-      if (mode is not None):
-         os.chmod(fp.fileno(), mode)
+   fp = open_(path, "wt")
+   ossafe(fp.write, "can't write: %s" % path, content)
+   if (mode is not None):
+      ossafe(os.chmod, "can't chmod 0%o: %s" % (mode, path))
+   fp.close()
+
+def grep_p(path, rx):
+   """Return True if file at path contains a line matching regular expression
+      rx, False if it does not."""
+   rx = re.compile(rx)
+   try:
+      with open(path, "rt") as fp:
+         for line in fp:
+            if (rx.search(line) is not None):
+               return True
+      return False
+   except OSError as x:
+      FATAL("error reading %s: %s" % (path, x.strerror))
+
+def log(*args, color=None, prefix="", **kwargs):
+   if (color is not None):
+      color_set(color, log_fp)
+   if (log_festoon):
+      prefix = ("%5d %s  %s"
+                % (os.getpid(),
+                   datetime.datetime.now().isoformat(timespec="milliseconds"),
+                   prefix))
+   print(prefix, file=log_fp, end="")
+   print(flush=True, file=log_fp, *args, **kwargs)
+   if (color is not None):
+      color_reset(log_fp)
+
+def log_setup(verbose_):
+   global verbose, log_festoon, log_fp
+   assert (0 <= verbose_ <= 2)
+   verbose = verbose_
+   if ("CH_LOG_FESTOON" in os.environ):
+      log_festoon = True
+   file_ = os.getenv("CH_LOG_FILE")
+   if (file_ is not None):
+      verbose = max(verbose_, 1)
+      log_fp = open_(file_, "at")
+   atexit.register(color_reset, log_fp)
+   DEBUG("verbose level: %d" % verbose)
 
 def mkdirs(path):
    DEBUG("ensuring directory: " + path)
-   os.makedirs(path, exist_ok=True)
+   try:
+      os.makedirs(path, exist_ok=True)
+   except OSError as x:
+      ch.FATAL("can't create directory: %s: %s: %s"
+               % (path, x.filename, x.strerror))
+
+def open_(path, mode, *args, **kwargs):
+   "Error-checking wrapper for open()."
+   return ossafe(open, "can't open for %s: %s" % (mode, path),
+                 path, mode, *args, **kwargs)
 
 def ossafe(f, msg, *args, **kwargs):
    """Call f with args and kwargs. Catch OSError and other problems and fail
       with a nice error message."""
    try:
-      f(*args, **kwargs)
+      return f(*args, **kwargs)
    except OSError as x:
       FATAL("%s: %s" % (msg, x.strerror))
 
 def rmtree(path):
    if (os.path.isdir(path)):
       DEBUG("deleting directory: " + path)
-      shutil.rmtree(path)
+      try:
+         shutil.rmtree(path)
+      except OSError as x:
+         ch.FATAL("can't recursively delete directory %s: %s: %s"
+                  % (path, x.filename, x.strerror))
    else:
       assert False, "unimplemented"
 
@@ -901,16 +959,19 @@ def storage_default():
       FATAL("can't get username: $USER not set")
    return "/var/tmp/%s/ch-grow" % username
 
-def symlink(target, source):
+def symlink(target, source, clobber=False):
+   if (clobber and os.path.isfile(source)):
+      unlink(source)
    try:
       os.symlink(target, source)
    except FileExistsError:
       if (not os.path.islink(source)):
-         FATAL("can't symlink: source exists and isn't a symlink: %s"
-               % source)
+         FATAL("can't symlink: source exists and isn't a symlink: %s" % source)
       if (os.readlink(source) != target):
          FATAL("can't symlink: %s exists; want target %s but existing is %s"
                % (source, target, os.readlink(source)))
+   except OSError as x:
+      ch.FATAL("can't symlink: %s -> %s: %s" % (source, target, x.strerror))
 
 def tree_child(tree, cname):
    """Locate a descendant subtree named cname using breadth-first search and
@@ -956,3 +1017,7 @@ def tree_terminals(tree, tname):
    for j in tree.children:
       if (isinstance(j, lark.lexer.Token) and j.type == tname):
          yield j.value
+
+def unlink(path, *args, **kwargs):
+   "Error-checking wrapper for os.unlink()."
+   ossafe(os.unlink, "can't unlink: %s" % path, path)
